#include "aladin.h"

static void reorder_bits(unsigned char *buf, int len) {
	int i;
	unsigned char j;

	for(i = 0; i < len; i++) {
		j  = (buf[i] & 0x01) << 7;
		j += (buf[i] & 0x02) << 5;
		j += (buf[i] & 0x04) << 3;
		j += (buf[i] & 0x08) << 1;
		j += (buf[i] & 0x10) >> 1;
		j += (buf[i] & 0x20) >> 3;
		j += (buf[i] & 0x40) >> 5;
		j += (buf[i] & 0x80) >> 7;
		buf[i] = j;
	}
	return;
}

static int checksum(unsigned char *buf) {
	int i;
	unsigned long file, comp;

	file = buf[0x7fd];
	file <<= 8;
	file += buf[0x7fc];

	for(i = 0, comp = 0; i < 0x7fc; i++) {
		comp += buf[i];
	}
	comp += 0x1fe;
	comp %= (1 << 16);

	return(file == comp);
}

LOG * read_aladin(char * tty) {
	int fd, i, rc, dev, t;
	char twirly[] = "/-\\|";
	unsigned long now;
	unsigned int rts = TIOCM_RTS;
	struct termios tio;
	unsigned char buf[2048];
	struct stat st_buf;

retry:
	if (stat(tty, &st_buf) < 0) {
		fprintf(stderr, "%s: ", tty);
		perror("stat");
		exit(1);
	}
	if (S_ISCHR(st_buf.st_mode)) {
		dev = 1;
	} else if (S_ISREG(st_buf.st_mode)) {
		dev = 0;
	} else {
		fprintf(stderr, "%s is neither a character device nor a file.\n", tty);
		exit(1);
	}
	if ((fd = open(tty, O_RDONLY)) < 0) {
		fprintf(stderr, "%s: ", tty);
		perror("open");
		exit(1);
	}

	if (dev) {
		tio.c_iflag = 0;
		tio.c_oflag = 0;
		tio.c_lflag = 0;
		tio.c_cflag = CS8|CLOCAL|CREAD;
		tio.c_cc[VMIN] = 1;
		tio.c_cc[VTIME] = 0;
		cfsetispeed(&tio, B19200);
		if (debug) {
			printf("Calling tcsetattr() on %s\n", tty);
			fflush(stdout);
		}
		if (tcsetattr(fd, TCSANOW, &tio) < 0) {
			fprintf(stderr, "%s: ", tty);
			perror("tcsetattr");
			exit(1);
		}
		if (debug) {
			printf("Calling ioctl() on %s\n", tty);
			fflush(stdout);
		}
		if (ioctl(fd, TIOCMBIC, &rts) < 0) {
			fprintf(stderr, "%s: ", tty);
			perror("ioctl");
			exit(1);
		}
		if (debug) {
			printf("Serial port setup complete.\n");
			fflush(stdout);
		}
	}

	for(i = 0, t = 0; i < 4; t++) {
		if (read(fd, buf, 1) < 0) {
			fprintf(stderr, "%s: ", tty);
			perror("read");
			exit(1);
		}
		if (i < 3) {
			if (buf[0] == 85) {
				i++;
			} else {
				if (buf[0] != 0) {
					fprintf(stderr, "%c (%02x).. Transmission failed. Reset state.\n", twirly[t%4], buf[0]);
				}
				i = 0;
			}
		} else if (buf[0] == 0) {
			i++;
		} else {
			fprintf(stderr, "%c (%02x).. Transmission failed. Reset state.\n", twirly[t%4], buf[0]);
			i = 0;
		}
	}
	if (dev) {
		fprintf(stderr, "Receiving data... ");
		fflush(stderr);
	}

	for(i = 0; i < 2046;) {
		rc = read(fd, buf+i, 2046-i);
		if (rc < 0) {
			fprintf(stderr, "%s: ", tty);
			perror("read");
			exit(1);
		}
		i += rc;
	}
	close(fd);

	if (dev) {
		fprintf(stderr, "done.\n");
	}

	if (debug >= 2) {
		printf("Dumping aladin raw data to \"dump\".\n");
		if ((fd = open("dump", O_WRONLY | O_CREAT | O_TRUNC, 0644)) < 0) {
			fprintf(stderr, "dump: ");
			perror("open");
			exit(1);
		}
		write(fd, "UUU\0", 4);
		write(fd, buf, 2046);
		close(fd);
	}

	reorder_bits(buf, 2046);

	if (!checksum(buf)) {
		fprintf(stderr, "Invalid checksum. ");
		if (dev) {
			fprintf(stderr, "Retry transmission.\n");
			goto retry;
		} else {
			fprintf(stderr, "Aborting.\n");
			exit(1);
		}
	}

	if (dev) {
		now = (unsigned long) time(NULL);
	} else {
		now = (unsigned long) st_buf.st_mtime;
	}

	if (debug >= 2) {
		dump_raw_data(buf, now);
	}

	return(reformat(buf, now));
}

